/*
 * Copyright (c) 1997 QUALCOMM Inc 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by QUALCOMM Inc
 * 4. Neither the name of QUALCOMM Inc nor the names of their
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRUSTEES AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TRUSTEES OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include <stdio.h>
#include <sys/param.h>
#include <lber.h>
#include <ldap.h>
#include "protos.h"
#include "ph2ldap.h"


#define PRINT_WIDTH 60

LDAP *ld = NULL;
char *LDAP_HOST=NULL;
char *LDAP_BASEDN=NULL;

int GIVEN_PORT = LDAP_PORT;
int LDAP_SEARCH_SCOPE = LDAP_SCOPE_SUBTREE;

/* default time and size limits */
int LDAP_SIZELIMIT = 0;
int LDAP_TIMELIMIT = 0;

extern FILE *Output;

int NUM_MAPPINGS       = 0;
int NUM_DENY_FIELDS    = 0;
int NUM_DEFAULT_FIELDS = 0;
int NUM_DEF_QUERY_FIELDS = 0;
int RETURN_ALL         = FALSE;

char *mappings[MAX_MAPPINGS];
char *deny_fields[MAX_MAPPINGS];
char *default_fields[MAX_MAPPINGS];
char *def_query_fields[MAX_MAPPINGS];


int InitLdap(char *ldap_host);
int ExitLdap();

int SearchLdap(char *filter, char **attrs);
int GetLdapConfig();
int MakeLdapFilter(ARG *argp, char *filter, char **fields);
int map_ph2ldap(char *entry, char *ldap_entry);
int map_ldap2ph(char *entry, char *ph_entry);

char * strip_line(char *str);
void fail(char *s);

void fail(char *s)
{
 fprintf(Output, "500: Unsuccessful due to: %s\n", s); 
 DoQuit(NULL);
}



/*
 *  Initializes LDAP connection.  Open the ldap connection and
 *  bind to it
 */
int InitLdap(char *ldap_host) 
{

     /* open a connection and authenticate */
     if ((ld = ldap_open(ldap_host, GIVEN_PORT)) == NULL)
	fail("Open");
    
     ld->ld_sizelimit = LDAP_SIZELIMIT; 
     ld->ld_timelimit = LDAP_TIMELIMIT; 

     if (ldap_bind_s(ld, NULL, NULL, LDAP_AUTH_SIMPLE) != LDAP_SUCCESS)
	fail("Bind");
     
    
     return 1;

} /* init */


/*
 *   Search ldap for the given filter and return the given attributes 
 *   or fields .  On success, this function sends the results to the
 *   client using DoReply function.  This is heart of the gateway as all
 *   the action is here.
 */

int SearchLdap(char *filter, char **attrs)
{
     LDAPMessage *entry, *results;
     char *a, *dn;
     char **v;
     int i, index;
     int count=0, num_values = 0, num_entries=0;
     int width=20;
     int status;
     int DONE = 0;
     int haveError = 0;
     char tfield[MAX_FILTER];

     /* search for entries, return all attrs */ 
     
     if( (attrs[0] == NULL) && NUM_DEFAULT_FIELDS && !RETURN_ALL)
          status = ldap_search_s(ld, LDAP_BASEDN, LDAP_SEARCH_SCOPE,
		filter, default_fields, 0, &results);
     else
          status = ldap_search_s(ld, LDAP_BASEDN, LDAP_SEARCH_SCOPE,
		filter, attrs, 0, &results);
     
     if(status != LDAP_SUCCESS)
     {
        char err[256];
	switch(status)
	{
	    case LDAP_FILTER_ERROR:
               DoReply(LR_ERROR, "%d:%s", status, ldap_err2string(status));
	       DoReply(LR_OK, "Ok.");
	       return 0;
	       break;
            case LDAP_SIZELIMIT_EXCEEDED:
               DoReply(-LR_TOOMANY, "Too many entries found.");
               break;
            case LDAP_TIMELIMIT_EXCEEDED:
               DoReply(-LR_ERROR, "Directory time limit exceeded.");
               break;

	    default:
                memset(err, 0, 256);
                /* sprintf(err, "Search with code %d", status); */
                sprintf(err, "%d:%s", status, ldap_err2string(status));
		fail(err);
		DoReply(-LR_OK, "LDAP Syntax error.");
		return -1;
	}
     }
     
     num_entries = ldap_count_entries(ld,  results);

     if(num_entries > 0)
     { 
	char buf[256];
	sprintf(buf, "There %s %d match%s to your query.",
				num_entries > 1 ? "were" : "was", 
				num_entries, 
				num_entries > 1 ? "es":"");
	DoReply(LR_NUMRET, buf);
     }

     /* step through each entry returned and print it to Ph client*/
     for (entry = ldap_first_entry(ld, results); entry != NULL; 
	    entry = ldap_next_entry(ld, entry)) 
     {
	 BerElement *berptr;
	 count++;
	 
	 
	/* step through each LDAP attribute and print its values */
	
	for (a = ldap_first_attribute(ld, entry, &berptr); a != NULL;
	    a = ldap_next_attribute(ld, entry, berptr)) 
	{
	      v = ldap_get_values(ld, entry, a);   
              num_values = ldap_count_values(v);
              if(num_values == 0)
                break;

              /*
               * if there is no access to ldap attributes, skip them here
               */

              if( !deny_access_to(a))
              { 

                map_ldap2ph(a, tfield);

                for(i=0; i < num_values; i++)
                {

		  int orig = strlen(v[i]);
		  int len = orig;
		  char value[256];
		  int first_time = 1;
		  char *ptr, *nl = v[i];
                  
                  while( (ptr = strstr(nl, "\\n")))
                  {
                     *ptr = 0;
                     len = len - strlen(nl)-2;
                     if( first_time)
		        DoReply(-LR_OK, "%d:%*s: %s", count, width, tfield, nl);
		     else
		        DoReply(-LR_OK, "%d:%*s: %s", count, width, "", nl);
		     nl = ptr+2; 
		     first_time = 0;
		  }

                  if(strlen(nl) > PRINT_WIDTH)
		  {
		     orig = len = strlen(nl);
	       	     do
		     {
		       memset(value, 0, 255);
		       strncpy(value, nl+orig-len, PRINT_WIDTH);

                       if( first_time)
		          DoReply(-LR_OK, "%d:%*s: %s", count, width, tfield, value);
		       else
		          DoReply(-LR_OK, "%d:%*s: %s", count, width, "", value);
		       first_time = 0;
		       len = len - PRINT_WIDTH;
		      }
		      while( len > 0);
                  }
 		  else
		  {
                       if( first_time)
		          DoReply(-LR_OK, "%d:%*s: %s", count, width, tfield, nl);
		       else
		          DoReply(-LR_OK, "%d:%*s: %s", count, width, "", nl);
		  }
	
                  
		/*

		  if( map_ldap2ph(a, tfield))
			DoReply(-LR_OK, "%d:%*s: %s", count, width, i==0 ? tfield : "", v[i]);
                  else
		        DoReply(-LR_OK, "%d:%*s: %s", count, width, i==0 ? a : "", v[i]);
		*/

                }

              }
	}

        /* if no value, no need to free */
        if(num_values)
	  ldap_value_free(v);

     } /* for each entry */
 
     if(!num_entries)
	DoReply(LR_NOMATCH, "No matches to your query.");
     else
	DoReply(LR_OK, "Ok."); 

}




/*
 * read the field config file.  Read into global FieldDescriptors
 */
int GetLdapConfig()
{
    int i;
    FILE    *fp;
    char    fconfig[MAXPATHLEN];
    char *ptr, *token, *first_token, *second_token;
    char buf[BUFLEN];
    int count = 0;  /* keep a count of counted mappings */
    int deny_index = 0;  /* keep a count of fields to deny access */
    int def_index = 0;  /* keep a count of default fields to return */
    int def_query_index = 0;  /* keep a count of query attributes to search by default */

    /* Database has the directory name for the default qi server directory */
    (void) sprintf(fconfig, "%s.cfg", Database);

    /* open the config file */
    if ((fp = fopen(fconfig, "r")) == NULL)
    {
	IssueMessage(LOG_ERR, "GetLdapConfig: fopen(%s): %s",
	    fconfig, strerror(errno));
	return (0);
    }


    while(fgets(buf, BUFLEN, fp) != NULL)
    {
       buf[strlen(buf)-1] = '\0';  /* remove the new line char */

       /* Find the LDAP host to which we must connect to */

       if( !strincmp(buf, HOST, strlen(HOST)))
       {
	 ptr = buf+strlen(HOST);
	 ptr = strip_line(ptr);
	 LDAP_HOST = strdup(ptr);
       }


       /* 
        *  Gather the base DN that is used for searching 
        *  Version 2 LDAP protocol does not allow to query for the defualt BAse DN 
        *  used by the server.  Ver 3 supposed to have this query capability, so 
        *  something to change for Ver 3 of LDAP
        */
       if( !strincmp(buf, BASE_DN, strlen(BASE_DN)))
       {
	 ptr = buf+strlen(BASE_DN);
	 ptr = strip_line(ptr);
	 LDAP_BASEDN = strdup(ptr);
       }
       
       /* 
        * get the size limit which is the maximum number of entries returned for 
        * each query. If this number is 0, then there is no size limit 
        */

       if( !strincmp(buf, SIZE_LIMIT, strlen(SIZE_LIMIT)))
       {
	 ptr = buf+strlen(SIZE_LIMIT);
	 ptr = strip_line(ptr);
	 LDAP_SIZELIMIT = atoi(ptr);
       }

       /* 
        * get the time limit which is the maximum number of seconds to spend for 
        * each query. If this number is 0, then there is no time limit 
        */
       if( !strincmp(buf, TIME_LIMIT, strlen(TIME_LIMIT)))
       {
	 ptr = buf+strlen(TIME_LIMIT);
	 ptr = strip_line(ptr);
	 LDAP_TIMELIMIT = atoi(ptr);
       }

       /* get the ldap port to use to connect to ldap servers*/
       if( !strincmp(buf, PORT, strlen(PORT)))
       {
	 ptr = buf+strlen(PORT);
	 ptr = strip_line(ptr);
	 GIVEN_PORT = atoi(ptr);
       }

       /* get the search scope for all ldap searches */
       if( !strincmp(buf, SEARCH_SCOPE, strlen(SEARCH_SCOPE)))
       {
	 ptr = buf+strlen(SEARCH_SCOPE);
	 ptr = strip_line(ptr);

         if(!strcmp(ptr, "LDAP_SCOPE_BASE"))
	 	LDAP_SEARCH_SCOPE = LDAP_SCOPE_BASE;
         else if(!strcmp(ptr, "LDAP_SCOPE_ONELEVEL"))
	 	LDAP_SEARCH_SCOPE = LDAP_SCOPE_ONELEVEL;
       }


       /* 
	*   gather ph to ldap and vice versa mappings here 
        *   syntax is:  MAP ldap_entry <space or tab> ph_entry 
        */

       if( !strincmp(buf, MAP, strlen(MAP)))
       {

	   ptr = buf+strlen(MAP);
	   ptr = strip_line(ptr);
	   first_token = strtok(ptr, " \t");  
	   if(first_token)
	      second_token = strtok(NULL, " \t\n");
	   if(second_token)
	   {
	       if( *first_token == '\"')
	       {
		 first_token++;
		 first_token[strlen(first_token)-1] = '\0';
	       }

	       if( *second_token == '\"')
	       {
		 second_token++;
		 second_token[strlen(second_token)-1] = '\0';
	       }
 
	       mappings[count++] = strdup(first_token); 
	       mappings[count++] = strdup(second_token); 
	   }
       }

       /* Gather DENY fields */
       if( !strincmp(buf, DENY, strlen(DENY)))
       {
		ptr = buf + strlen(DENY);
                ptr = strip_line(ptr);
                first_token = strtok(ptr, " ,\n");
                if(first_token)
                {
			deny_fields[deny_index++] = strdup(strip_line(first_token));
                }	
                while( token = strtok(NULL, " ,\n"))
			deny_fields[deny_index++] = strdup(strip_line(token));
                       
       }
       
       /* Gather default query attributes */
       if( !strincmp(buf, DEFAULT_QUERY, strlen(DEFAULT_QUERY)))
       {
		ptr = buf + strlen(DEFAULT_QUERY);
                ptr = strip_line(ptr);
                first_token = strtok(ptr, " ,\n");
                if(first_token)
                {
	            strip_line(first_token);
		    if( strcmp(first_token, "cn"))
		       def_query_fields[def_query_index++] = strdup(first_token);
                }	
                while( token = strtok(NULL, " ,\n"))
		{
		    strip_line(token);
		    if( strcmp(token, "cn"))
			def_query_fields[def_query_index++] = strdup(token);
		}
                       
       }


       /* Gather DEFAULT fields settings */
       if( !strincmp(buf, DEFAULT, strlen(DEFAULT)))
       {
		ptr = buf + strlen(DEFAULT);
                ptr = strip_line(ptr);
                first_token = strtok(ptr, " ,\n");
                if(first_token)
                {
			default_fields[def_index++] = strdup(strip_line(first_token));
                }	
                while( token = strtok(NULL, " ,\n"))
			default_fields[def_index++] = strdup(strip_line(token));
       }
                       
    } /* end of while == finish file processing */        

    NUM_MAPPINGS = count;
    NUM_DENY_FIELDS = deny_index;
    NUM_DEFAULT_FIELDS = def_index;
    NUM_DEF_QUERY_FIELDS = def_query_index;

    /* make the last one null to terminate the array */ 
    deny_fields[deny_index] = 0;
    default_fields[def_index] = 0;
    def_query_fields[def_query_index] = 0;

}





/*
 * Get rid of trailing spaces (and newlines and tabs ) as well as
 * beginning ones.  This routine just truncates the line in place.
 */

char * strip_line(char *str)
{
	register char *index;

	index = str;
	while (*index)
		index++;
	while (isspace(*--index))
	if (index < str)
			break;
	index[1] = '\0';
	while (isspace(*str))
	str++;
	return (str);
}



int MakeWordSearchFilter(char *filter, char *name, char *value)
{
    sprintf(filter, "(|(|(| (%s=%s) (%s=* %s *)) (%s=%s *)) (%s=* %s))",
	           name, value, name, value, name, value, name, value);	
}

int ExtendFilter(char *filter, char *name, char *value, char filter_type)
{
     char tfilter[MAX_FILTER];
     strcpy(tfilter, filter);
    
     sprintf(filter, "(%c%s(%s=%s))", filter_type, tfilter, name, value);
}
   


/*
 * Decide if a query request is valid returns the number of key fields found
 */
int 
MakeLdapFilter(ARG *argp, char *filter, char **fields)
{
        int i, count=0, ret_fields = 0, first_cn=1;
        int haveError = 0, sn;
        char filter_comp[MAX_FILTER];
        char tfilter[MAX_FILTER];
        char name[BUFLEN];
        char *ptr;

        /* reset return_all condition */ 
        RETURN_ALL = FALSE;
        memset(filter, 0, MAX_FILTER);
	argp = argp->aNext;	/* skip command name */

	/* collect arguments before ``return'' token */
	for (; argp; count++, argp = argp->aNext)
	{
		switch (argp->aType)
		{
    		    case RETURN:
                        ret_fields = 1;
			goto keyEnd; 
                    break;

		    case EQUAL:
			DoReply(-LR_SYNTAX, "=:No field or value specified.");
                        haveError = 1;
			break;

/*
		    case VALUE | EQUAL:
			argp->aSecond = strdup("");
			goto canonical;
*/

		    case VALUE:
			argp->aSecond = argp->aFirst;
			/* fall-through is deliberate */
		    case EQUAL | VALUE2:
#ifndef DEFQUERY
#define DEFQUERY "cn"		/* which field searched by default */
#endif /*DEFQUERY*/
			argp->aFirst = strdup(DEFQUERY);
			argp->aType = VALUE | EQUAL | VALUE2;
			/* again, we _should_ fall through here */
      canonical:

		    case VALUE | EQUAL | VALUE2:
                      
                        /* convert ph field into a ldap attribute based on the mappings */
                        /* use tfilter as temporary storage in which the conversion takes place*/

                        map_ph2ldap(argp->aFirst, tfilter);

                        /*
			 *  construct the ldap filter
 			 *  example:  (&(cn=sudheer)(mail=skoganti)) 
                         */

                        memset(filter_comp, 0, MAX_FILTER);
                        if( !strcmp(tfilter , "cn"))
                        {
			   MakeWordSearchFilter(filter_comp, tfilter, argp->aSecond);
			   for(i=0; i<NUM_DEF_QUERY_FIELDS; i++)
			      ExtendFilter(filter_comp, def_query_fields[i], argp->aSecond, LDAP_OR);
                        }
                        else 
                        {
		           if( strchr(argp->aSecond, '*'))
		              sprintf(filter_comp, "(%s=%s)", tfilter, argp->aSecond);
		           else
		              MakeWordSearchFilter(filter_comp, tfilter, argp->aSecond);
                        }

                        /*build the top filter */
                        if( strlen(filter) == 0 )
                        {
			    strcpy(filter, filter_comp);
			}
                        else
                        {
                          strcpy(tfilter, filter);
                          memset(filter, 0, MAX_FILTER);
                          sprintf(filter, "(&%s%s)", tfilter, filter_comp);
                        }
                        count++;  /* number of filters */
			break;

		    default:
			DoReply(-LR_ERROR, "Argument %d:parsing error.", count);
			haveError = 1;
			break;
		}
	} /* end for loop */

        keyEnd:
	   
	   if( !count)
           {
		haveError = 1;
                DoReply(-LR_NOKEY, "no non-null key field in query.");
           }
           if( ret_fields)
           {
		   /* get rid of return token */
		   argp=argp->aNext;

                   /*
                    *  copy each return field token into a table of strings
                    *  called fields which will be used in ldap search to
                    *  return only these particular fields.
                    */
		   for(i=0;argp; argp=argp->aNext, i++)
		   {

                       if(!strincmp(argp->aFirst, "all", 3))
                       { 
                         RETURN_ALL = TRUE;
                         break;
                       }

    		       /* truncate at the comma if one is there */
                       if(ptr = strstr(argp->aFirst, ","))   
                           *ptr = 0;

           	       /*  strip the fields of comma and convert to ldap attributes */
                       map_ph2ldap(argp->aFirst, tfilter);

		       fields[i] = strdup(tfilter);
		   } 

                   /* make last field NULL to recognize end of the array */
                   fields[i] = NULL;
           }
           else
             fields[0] = NULL;
           		

	return (haveError ? 0 : 1);
}

/*
 *  map ph entry to ldap attribute if there is a
 *  mapping defined for it 
 */

int map_ph2ldap(char *entry, char *ldap_entry)
{
      int i;
      for(i=1; i < NUM_MAPPINGS; i=i+2)
      {
	if( !stricmp(entry, mappings[i]))
	{
             strcpy(ldap_entry, mappings[i-1]);
             return TRUE;
	}
      }
      strcpy(ldap_entry, entry);
      return FALSE;
}



/*
 *  map ldap attribute to Ph field if there is a
 *  mapping defined for it 
 */

int map_ldap2ph(char *entry, char *ph_entry)
{
      int i;
      for(i=0; i < NUM_MAPPINGS; i=i+2)
      {
	if( !stricmp(entry, mappings[i]))
	{
             strcpy(ph_entry, mappings[i+1]);
             return TRUE;
	}
      }
      strcpy(ph_entry, entry);
      return FALSE;
}






int deny_access_to(char *ldap_entry)
{
      int i;
      for(i=0; i < NUM_DENY_FIELDS; i++)
      {
	if( !stricmp(ldap_entry, deny_fields[i]))
             return TRUE;
      }
      return FALSE;
}



int ExitLdap()
{
  if(ld)
   ldap_unbind(ld);
}
